-------------------------------------------------------------------------------
-- Locals                                                                    --
-------------------------------------------------------------------------------

local L = AceLibrary("AceLocale-2.2"):new("ClosetGnome")
local dewdrop = AceLibrary("Dewdrop-2.0")
local tablet = AceLibrary("Tablet-2.0")

local defaultIcon = "Interface\\Icons\\INV_Chest_Cloth_17"

-- GRRRR! We shouldn't need this variable, but BLIZZARD IS ON CRACK.
local createOrUpdateSetName = nil

local deequipQueue = nil
local queuedSet = {}
local lastEquippedSet = nil
local slotLocks = nil
local queuedForDelete = nil

local activeSlots = nil

local keybindingFrame = "ClosetGnomeKeybindingFrame"

local _G = getfenv(0)

local slots = {
	[1] = {name = "Head"},
	[3] = {name = "Shoulder"},
	[5] = {name = "Chest"},
	[6] = {name = "Waist"},
	[7] = {name = "Legs"},
	[8] = {name = "Feet"},
	[9] = {name = "Wrist"},
	[10] = {name = "Hands"},
	[17] = {name = "SecondaryHand"},
	[16] = {name = "MainHand"},
	[18] = {name = "Ranged"},
	[0] = {name = "Ammo"},
	[2] = {name = "Neck"},
	[15] = {name = "Back"},
	[11] = {name = "Finger0"},
	[12] = {name = "Finger1"},
	[13] = {name = "Trinket0"},
	[14] = {name = "Trinket1"},
	[19] = {name = "Tabard", inactive = true},
	[4] = {name = "Shirt", inactive = true},
}

local NUM_QUIPS = 35

if (GetLocale() == "zhTW") then
	NUM_QUIPS = 3
end

local PickupContainerItem = _G.PickupContainerItem
local AutoEquipCursorItem = _G.AutoEquipCursorItem
local EquipCursorItem = _G.EquipCursorItem
local PickupInventoryItem = _G.PickupInventoryItem

-------------------------------------------------------------------------------
-- Addon declaration                                                         --
-------------------------------------------------------------------------------

ClosetGnome = AceLibrary("AceAddon-2.0"):new("AceHook-2.1", "AceConsole-2.0", "AceDB-2.0", "AceEvent-2.0", "FuBarPlugin-2.0")
local ClosetGnome = ClosetGnome

ClosetGnome.revision = tonumber(string.sub("$Revision: 40867 $", 12, -3))

local function IsEmpty()
	return next(ClosetGnome.OnMenuRequest.args.wear.args) == nil
end

local options = {
	type = "group",
	name = L["ClosetGnome"],
	desc = L["ClosetGnome options."],
	args = {
		add = {
			type = "execute",
			name = L["Add"],
			desc = L["Creates a new set, or updates an existing one."],
			func = "CreateOrUpdateSet",
			order = 101,
		},
		wear = {
			type = "group",
			name = L["Wear"],
			desc = L["Change equipment set."],
			pass = true,
			func = function(key)
				dewdrop:Close()
				if IsShiftKeyDown() then
					ClosetGnome:CreateOrUpdateSet(key)
				else
					ClosetGnome:WearSet(key)
				end
			end,
			args = {},
			order = 102,
			disabled = IsEmpty
		},
		key = {
			type = "group",
			name = L["Keybinding"],
			desc = L["Assign a keybinding to a set."],
			pass = true,
			get = function(key) return ClosetGnome.db.char.keybindings[key] end,
			set = function(key, value)
				if value == nil then
					ClosetGnome:ClearBinding(key)
				else
					ClosetGnome:RegisterBinding(value, key)
				end
			end,
			args = {},
			order = 103,
			disabled = IsEmpty
		},
		spacer = { type = "header", order = 104 },
		alwaysWeapons = {
			type = "toggle",
			name = L["Always equip weapons"],
			desc = L["Equip weapons in sets even if you are in combat."],
			get = function() return ClosetGnome.db.char.alwaysWeapons end,
			set = function(v) ClosetGnome.db.char.alwaysWeapons = v end,
			order = 105,
		},
		quips = {
			type = "toggle",
			name = L["Quips"],
			desc = L["Toggle outputting random quips when equipping sets."],
			get = function() return ClosetGnome.db.profile.quips end,
			set = function(v) ClosetGnome.db.profile.quips = v end,
			order = 106,
		},
		delete = {
			type = "group",
			name = L["Delete"],
			desc = L["Delete a set."],
			pass = true,
			func = "DeleteSet",
			args = {},
			order = 107,
			disabled = IsEmpty
		},
	},
}

-- FuBar stuff.
ClosetGnome.hideWithoutStandby = true
ClosetGnome.hasIcon = defaultIcon
ClosetGnome.hasNoColor = true
ClosetGnome.defaultMinimapPosition = 240
ClosetGnome.clickableTooltip  = true
ClosetGnome.OnMenuRequest = options

-------------------------------------------------------------------------------
-- Stack                                                                     --
-------------------------------------------------------------------------------
--[[
local function newStack(func)
	local x = {}
	local stack = {}
	function x:Push(id) table.insert(stack, 1, id) end
	function x:Pop()
		if #stack == 0 then return nil end
		if type(func) == "function" then
			func(table.remove(stack, 1))
		else
			ClosetGnome[func](ClosetGnome, table.remove(stack, 1))
		end
	end
	return x
end]]

-- newStack(function(item)
--   ChatFrame1:AddMessage(tostring(item))
-- end)
-- x:Push("test")
-- x:Push("test2")
-- x:Pop()
-- x:Pop()
-- =
-- test2
-- test

-------------------------------------------------------------------------------
-- Initialization                                                            --
-------------------------------------------------------------------------------

function ClosetGnome:OnInitialize()
	self.version = (self.version or "1").."."..tostring(self.revision or "1")

	_G.CLOSETGNOME_VERSION = self.version
	_G.CLOSETGNOME_REVISION = self.revision

	self:RegisterDB("ClosetGnomeDB", "ClosetGnomePerCharDB")
	self:RegisterDefaults("profile", {
		quips = true,
		hideBetaNotice = false,
	})
	self:RegisterDefaults("char", {
		set = {},
		icons = {},
		keybindings = {},
		alwaysWeapons = true,
	})

	self:RegisterChatCommand("/closetgnome", "/cg", options, "CLOSETGNOME")
end

function ClosetGnome:OnEnable()
	if type(_G.StaticPopupDialogs) ~= "table" then
		_G.StaticPopupDialogs = {}
	end
	if not _G.StaticPopupDialogs["ClosetGnomeAdd"] then
		_G.StaticPopupDialogs["ClosetGnomeAdd"] = {
			text = L["|cffeda55fLeft-Click|r a slot to toggle it for this set. Green slots are |cff00ff00enabled|r, red are |cffff0000disabled|r.\n\nIf you want to assign an |cff0000fficon|r to this set, you can |cffeda55fCtrl-Click|r a slot to use that slots item as the icon.\n\nType the desired set name below and click Add when you are done."],
			button1 = L["Add"],
			button2 = L["Cancel"],
			OnCancel = function() self:CancelCreateSet() end,
			sound = "levelup2",
			whileDead = 1,
			hideOnEscape = 1,
			timeout = 0,
			OnShow = function()
				-- We have to do this onshow to reset the previous text
				_G[this:GetName().."EditBox"]:SetText(createOrUpdateSetName or "")
			end,
			OnHide = function()
				_G[this:GetName().."EditBox"]:SetText("")
			end,
			OnAccept = function()
				local name = _G[this:GetParent():GetName().."EditBox"]:GetText()
				self:AddSetFromDoll(name)
				createOrUpdateSetName = nil
			end,
			hasEditBox = 1,
		}
	end
	if not _G.StaticPopupDialogs["ClosetGnomeDelete"] then
		_G.StaticPopupDialogs["ClosetGnomeDelete"] = {
			text = L["Are you sure you want to delete the set %s?"],
			button1 = L["Delete"],
			button2 = L["Cancel"],
			showAlert = 1,
			timeout = 0,
			OnAccept = function() self:DeleteSet(queuedForDelete, true) end,
			OnCancel = function() queuedForDelete = nil end,
		}
	end

	-- Create the button frames for keybindings
	for k, v in pairs(self.db.char.keybindings) do
		self:RegisterBinding(v, k)
	end

	if AceLibrary("AceEvent-2.0"):IsFullyInitialized() then
		self:AceEvent_FullyInitialized()
	end
	self:RegisterEvent("AceEvent_FullyInitialized")

	self:UpdateSetMenu()
end

local function reclaim(t)
	if type(t) ~= "table" then return end
	for k, v in pairs(t) do
		if type(v) == "table" then
			t[k] = reclaim(v)
		else
			t[k] = nil
		end
	end
	t = nil
	return nil
end

function ClosetGnome:OnDisable()
	activeSlots = reclaim(activeSlots)
	slotLocks = reclaim(slotLocks)

	for k, v in pairs(self.db.char.keybindings) do
		self:UnsetBinding(k)
	end
end


-------------------------------------------------------------------------------
-- Keybindings                                                               --
-------------------------------------------------------------------------------

local function ValidateBinding(combo)
	return AceLibrary("AceConsole-2.0").keybindingValidateFunc(combo)
end

local function KeybindingClicked()
	if type(this.set) ~= "string" or not ValidateBinding(this.combo) then return end
	ClosetGnome:WearSet(this.set)
end

function ClosetGnome:RegisterBinding(combo, set)
	if not combo or not set then return end

	if not ValidateBinding(combo) then
		self:Print(L["%s is not a valid keybinding."]:format("|cffd9d919"..combo.."|r"))
		return
	end

	local shouldPrint = true
	for k, v in pairs(self.db.char.keybindings) do
		if v == combo then
			shouldPrint = nil
		elseif k == set then
			self:UnsetBinding(set)
		end
	end

	if shouldPrint then
		self:Print(L["Registering keybinding %s to set %s."]:format("|cffd9d919"..combo.."|r", "|cffd9d919"..set.."|r"))
	end

	local frameName = keybindingFrame..combo
	local frame = _G[frameName] or CreateFrame("Button", frameName, _G["UIParent"])
	frame.combo = combo
	frame.set = set
	frame:SetScript("OnMouseUp", KeybindingClicked)

	self.db.char.keybindings[set] = combo
	SetBindingClick(combo, frameName)

	self:UpdateDisplay()
end

function ClosetGnome:UnsetBinding(set)
	if not set or not self.db.char.keybindings[set] then return end
	local combo = self.db.char.keybindings[set]

	local frame = _G[keybindingFrame..combo]
	if frame then
		frame:SetScript("OnMouseUp", nil)
		frame = nil
	end

	SetBinding(combo) -- Unset the binding.
end

function ClosetGnome:ClearBinding(set)
	if not set or not self.db.char.keybindings[set] then return end

	self:UnsetBinding(set)

	self:Print(L["Removing keybinding %s from set %s."]:format("|cffd9d919"..self.db.char.keybindings[set].."|r", "|cffd9d919"..set.."|r"))
	self.db.char.keybindings[set] = nil

	self:UpdateDisplay()
end

-------------------------------------------------------------------------------
-- Menu                                                                      --
-------------------------------------------------------------------------------

function ClosetGnome:UpdateSetMenu()
	local opt = self.OnMenuRequest.args

	opt.wear.args = reclaim(opt.wear.args)
	opt.delete.args = reclaim(opt.delete.args)
	opt.key.args = reclaim(opt.key.args)

	opt.wear.args = {}
	opt.delete.args = {}
	opt.key.args = {}

	for k, v in pairs(self.db.char.set) do
		local icon = self.db.char.icons[k]
		if not self:HideSetFromUI(k) then
			opt.wear.args[k] = {
				type = "execute",
				name = k,
				desc = L["Equip %s, or hold down Shift to edit this set."]:format(k),
				icon = icon,
			}
		end
		opt.delete.args[k] = {
			type = "execute",
			name = k,
			desc = L["Deletes the equipment set %s."]:format(k),
			icon = icon,
		}
		opt.key.args[k] = {
			type = "text",
			name = k,
			desc = L["Assign a keybinding to %s (or empty to clear)."]:format(k),
			validate = "keybinding",
			icon = icon,
		}
	end
end

-------------------------------------------------------------------------------
-- Addon methods                                                             --
-------------------------------------------------------------------------------

local function OnEnter(...)
	if type(_G.PaperDollItemSlotButton_OnEnter) == "function" then
		this = this:GetParent()
		_G.PaperDollItemSlotButton_OnEnter(...)
	end
end

local function OnLeave(...)
	this:GetParent().updateTooltip = nil
	GameTooltip:Hide()
	ResetCursor()
end

local function OnClick(...)
	local id = this.id

	if IsShiftKeyDown() or (not IsModifierKeyDown() and CursorHasItem()) and type(PaperDollItemSlotButton_OnClick) == "function" then
		this = this:GetParent()
		_G.PaperDollItemSlotButton_OnClick("LeftButton")
	elseif IsControlKeyDown() then
		-- Don't want an empty slot as icon.
		local currentItem = ClosetGnome:ItemNameFromSlot(id)
		if currentItem == false then return end

		if activeSlots["icon"] then
			local oldSlot = activeSlots["icon"]
			activeSlots[oldSlot] = true
			_G["ClosetGnomeTexture"..oldSlot]:SetVertexColor(0, 255, 0)
		end
		activeSlots["icon"] = id
		activeSlots[id] = true
		this.texture:SetVertexColor(0, 0, 255)
	else
		activeSlots[id] = not activeSlots[id]
		if activeSlots[id] then
			if activeSlots["icon"] == id then
				this.texture:SetVertexColor(0, 0, 255)
			else
				this.texture:SetVertexColor(0, 255, 0)
			end
		else
			this.texture:SetVertexColor(255, 0, 0)
		end
	end
end

-- Opens the character frame and lets the user uncheck the items he doesn't
-- want.
function ClosetGnome:CreateOrUpdateSet(input)
	if UnitAffectingCombat("player") or InCombatLockdown() then return end

	activeSlots = {}
	for id, slot in pairs(slots) do
		local gslot = _G["Character" .. slot.name .. "Slot"]
		if not slot.frame and gslot then
			-- We used to just :GetScript("OnClick") on the existing slot,
			-- and save that so that we could override it and put it back
			-- after we close the character frame, but this is not possible
			-- in TBC with the protected code stuff. WoW will complain that
			-- the script is tainted, and the normal slots won't work again
			-- after adding a set, so we create our own frame that we set on
			-- top of the existing one now.
			local button = CreateFrame("Button", "ClosetGnome"..id, gslot)
			button:SetAlpha(0)
			button:SetHeight(gslot:GetHeight())
			button:SetWidth(gslot:GetWidth())
			button:SetPoint("CENTER", gslot, "CENTER", 0, 1)
			button:SetScript("OnEnter", OnEnter)
			button:SetScript("OnLeave", OnLeave)
			button:SetScript("OnClick", OnClick)
			button.id = id
			button:Hide()

			local texture = gslot:CreateTexture("ClosetGnomeTexture"..id, "OVERLAY")
			texture:SetTexture("Interface\\Buttons\\UI-ActionButton-Border")
			texture:SetVertexColor(0, 255, 0)
			texture:SetBlendMode("ADD")
			texture:SetAlpha(0.75)
			texture:SetHeight(68)
			texture:SetWidth(68)
			texture:SetPoint("CENTER", gslot, "CENTER", 0, 1)
			texture:Hide()

			button.texture = texture

			slot.frame = button
			slot.texture = texture
		end
		if slot.frame and gslot then
			if slot.inactive then
				activeSlots[id] = false
				slot.texture:SetVertexColor(255, 0, 0)
			else
				activeSlots[id] = true
				slot.texture:SetVertexColor(0, 255, 0)
			end
			slot.frame:EnableMouse(true)
			slot.frame:Show()
			slot.texture:Show()
		end
	end

	-- We're updating a set!
	if type(input) == "string" and self.db.char.set[input] then
		self:WearSet(input)
		local set = self.db.char.set[input]
		for k in pairs(activeSlots) do
			if set[k] ~= nil then -- Can also be false, so check for nil.
				activeSlots[k] = true
				slots[k].texture:SetVertexColor(0, 255, 0)
			else
				activeSlots[k] = false
				slots[k].texture:SetVertexColor(255, 0, 0)
			end
		end
	end

	-- Hook Fizzle to prevent it from styling the slots while we add a set.
	if IsAddOnLoaded("Fizzle") and type(_G.Fizzle) == "table" and type(_G.Fizzle.UpdateItems) == "function" then
		if type(_G.Fizzle.HideBorders) == "function" then
			_G.Fizzle:HideBorders()
		end
		self:Hook(_G.Fizzle, "UpdateItems", "Noop")
	end

	_G.PanelTemplates_SetTab(_G.CharacterFrame, _G.PaperDollFrame:GetID())
	if _G.CharacterFrame:IsVisible() then
		if not _G.PaperDollFrame:IsVisible() then
			_G.CharacterFrame_ShowSubFrame("PaperDollFrame")
		end
	else
		_G.ShowUIPanel(_G.CharacterFrame)
		_G.CharacterFrame_ShowSubFrame("PaperDollFrame")
	end

	_G.OpenAllBags(1)

	self:Hook("CharacterFrame_OnHide", true)
	self:Hook("CharacterFrameTab_OnClick", "Noop", true)
	self:Hook("PaperDollItemSlotButton_OnModifiedClick", "Noop", true)

	dewdrop:Close()

	createOrUpdateSetName = input
	_G.StaticPopup_Show("ClosetGnomeAdd")
end

function ClosetGnome:CancelCreateSet()
	self:CharacterFrame_OnHide()
end

function ClosetGnome:AddSetFromDoll(name)
	local slots = {}
	local icon = nil
	for slot, active in pairs(activeSlots) do
		if slot == "icon" then
			icon = active
		elseif active then
			table.insert(slots, slot)
		end
	end
	self:AddSet(name, slots, icon)
	activeSlots = reclaim(activeSlots)
	slots = reclaim(slots)
	icon = nil

	self:CharacterFrame_OnHide()
end

function ClosetGnome:AddSet(name, slots, iconSlot)
	local concatedName = name:gsub(" ", "")
	if not name or name == "" or concatedName == "" then
		self:Print(L["Please use a proper name for your set."])
		return
	end
	if self:HasSet(concatedName) then name = concatedName end

	local text = L["Added set: %s."]:format(name)
	local event = "ClosetGnome_AddSet"
	if self:HasSet(name) then
		text = L["Updating set: %s."]:format(name)
		event = "ClosetGnome_UpdateSet"
		for k in pairs(self.db.char.set[name]) do
			self.db.char.set[name][k] = nil
		end
	end

	for i, slot in pairs(slots) do
		local currentItem = self:ItemNameFromSlot(slot)
		if not self.db.char.set[name] then self.db.char.set[name] = {} end
		self.db.char.set[name][slot] = currentItem
	end
	if not self.db.char.set[name] then return end

	if iconSlot then
		self.db.char.icons[name] = GetInventoryItemTexture("player", iconSlot)
	end

	lastEquippedSet = name
	self:TriggerEvent(event, name)
	self:Print(text)

	self:UpdateSetMenu()
	self:UpdateDisplay()
end

function ClosetGnome:DeleteSet(name, verified)
	if not verified and not IsShiftKeyDown() then
		queuedForDelete = name
		StaticPopup_Show("ClosetGnomeDelete", name)
		return
	end

	queuedForDelete = nil
	if self.db.char.set[name] then
		self.db.char.set[name] = reclaim(self.db.char.set[name])
		self.db.char.icons[name] = nil
		self:ClearBinding(name)

		if lastEquippedSet == name then
			lastEquippedSet = nil
		end

		self:TriggerEvent("ClosetGnome_DeleteSet", name)
		self:Print(L["Deleted set: %s."]:format(name))
		self:UpdateSetMenu()
		self:UpdateDisplay()
		dewdrop:Refresh(1)
		dewdrop:Refresh(2)
	end
end

local function GetOppositeSlot(slot)
	if type(slot) ~= "number" then error("Only takes numbers.") end
	if slot == 11 then return 12
	elseif slot == 12 then return 11
	elseif slot == 13 then return 14
	elseif slot == 14 then return 13
	elseif slot == 16 then return 17
	elseif slot == 17 then return 16
	end
	return nil
end

function ClosetGnome:ProcessDeequipQueue()
	if not deequipQueue then return end
	if not slotLocks then slotLocks = {} end
	for i, slot in pairs(deequipQueue) do
		PickupInventoryItem(slot)
		local toBag, toSlot = ClosetGnome:LocateFreeSlot()
		if toBag ~= nil then
			slotLocks[toBag][toSlot] = true
			PickupContainerItem(toBag, toSlot)
		else
			AutoEquipCursorItem()
			break
		end
	end
	deequipQueue = reclaim(deequipQueue)
	slotLocks = reclaim(slotLocks)
end

-- ERR_2HANDED_EQUIPPED

local function ReallyWearSet(set, inCombat)
	local slotItems = {}
	for slot, item in pairs(set) do
		if not slotItems[slot] then slotItems[slot] = ClosetGnome:ItemNameFromSlot(slot) end
		if (not inCombat and slot > 10 and slot < 15) or slot == 16 or slot == 17 or slot == 18 then
			local oppositeSlot = GetOppositeSlot(slot)
			if oppositeSlot and not slotItems[oppositeSlot] then slotItems[oppositeSlot] = ClosetGnome:ItemNameFromSlot(oppositeSlot) end
			local swappable = true
			if (slot == 16 or slot == 17) and slotItems[slot] then
				local loc = ClosetGnome:GetItemEquipLoc(slotItems[slot])
				local oppositeLoc = slotItems[oppositeSlot] and ClosetGnome:GetItemEquipLoc(slotItems[oppositeSlot]) or nil
				swappable = loc == "INVTYPE_WEAPON" and (oppositeLoc and oppositeLoc == "INVTYPE_WEAPON" or true)
			end
			-- The item in this slot is not the one we want.
			if slotItems[slot] ~= item and swappable then
				if oppositeSlot and slotItems[oppositeSlot] == item then
					-- The item is in the opposite slot, swap it.
					PickupInventoryItem(slot)
					PickupInventoryItem(oppositeSlot)
				else
					ClosetGnome:EquipItem(slot, item)
				end
			elseif item ~= slotItems[slot] then
				ClosetGnome:EquipItem(slot, item)
			end
		elseif not inCombat and (slot < 11 or slot > 14) then
			if item ~= slotItems[slot] then
				ClosetGnome:EquipItem(slot, item)
			end
		end
	end
	slotItems = reclaim(slotItems)
end

function ClosetGnome:WearSet(...)
	ClearCursor()
	if CursorHasItem() or CursorHasSpell() or (UnitIsDeadOrGhost("player") and not UnitIsFeignDeath("player")) then return false end

	local size = select("#", ...)
	if size < 1 then return false end

	local setToWear = nil
	local name = nil

	if size == 1 then
		name = select(1, ...)
		setToWear = self.db.char.set[name]
	else
		for i = 1, size do
			local set = select(i, ...)
			if type(set) ~= "string" or not self.db.char.set[set] then return false end
		end
		setToWear = {}
		for i = 1, size do
			name = select(i, ...)
			for k, v in pairs(self.db.char.set[name]) do
				setToWear[k] = v
			end
		end
	end
	if type(setToWear) ~= "table" then return false end

	lastEquippedSet = name

	if self:IsSetFullyEquipped(setToWear) then
		return true
	end

	if UnitAffectingCombat("player") then
		if self.db.char.alwaysWeapons then
			deequipQueue = reclaim(deequipQueue)
			ReallyWearSet(setToWear, true)
			self:ProcessDeequipQueue()
			self:UpdateDisplay()

			if self:IsSetFullyEquipped(setToWear) then
				self:TriggerEvent("ClosetGnome_WearSet", name)
				if size > 1 then
					setToWear = reclaim(setToWear)
				end
				return true
			end
		end

		self:TriggerEvent("ClosetGnome_PartlyWearSet", name)
		if not self:IsEventRegistered("PLAYER_REGEN_ENABLED") then
			self:RegisterEvent("PLAYER_REGEN_ENABLED")
		end
		self:Print(L["In combat: %s queued for swap."]:format(name))

		if size > 1 then
			setToWear = reclaim(setToWear)
		end

		for i = 1, size do
			table.insert(queuedSet, select(i, ...))
		end
		return true
	end

	if self:IsEventRegistered("PLAYER_REGEN_ENABLED") then
		self:UnregisterEvent("PLAYER_REGEN_ENABLED")
	end

	if self.db.profile.quips then
		-- I don't want the addon name in front for this one.
		-- Thanks to MrPlow for the color :)
		DEFAULT_CHAT_FRAME:AddMessage(self:GetQuip(name))
	end

	deequipQueue = reclaim(deequipQueue)
	ReallyWearSet(setToWear, nil)
	self:ProcessDeequipQueue()
	self:TriggerEvent("ClosetGnome_WearSet", name)

	self:UpdateDisplay()

	if size > 1 then
		setToWear = reclaim(setToWear)
	end

	if type(queuedSet) == "table" then
		for i in ipairs(queuedSet) do
			table.remove(queuedSet, i)
		end
	end

	return true
end

-------------------------------------------------------------------------------
-- Utility                                                                   --
-------------------------------------------------------------------------------

function ClosetGnome:GetQuip(input)
	local rand = 1
	repeat rand = math.random(1, NUM_QUIPS) until L[rand] ~= nil
	return "|cffeda55f"..L[rand]:format("|cffd9d919"..input.."|cffeda55f").."|r"
end

function ClosetGnome:ItemNameFromSlot(slot)
	assert(slot, "Slot must be non-nil.")
	local slotId = tonumber(slot)
	if not slotId then slotId = GetInventorySlotInfo(slot) end
	if not slotId then error(string.format("Unknown inventory slot %q.", slot)) end
	local itemLink = GetInventoryItemLink("player", slotId)
	if not itemLink then return false end
	return itemLink
end

function ClosetGnome:IsSetFullyEquipped(input)
	local set = nil
	if type(input) == "string" then
		set = self.db.char.set[input]
	elseif type(input) == "table" then
		set = input
	end
	if not set then return false end
	for slot, item in pairs(set) do
		local currentItem = self:ItemNameFromSlot(slot)
		if item ~= currentItem then return false end
	end
	return true
end

function ClosetGnome:GetItemEquipLoc(link)
	return select(9, GetItemInfo(link))
end

function ClosetGnome:GetItemSubtype(link)
	return select(7, GetItemInfo(link))
end

function ClosetGnome:IsNormalBag(bagId)
	if bagId == 0 or bagId == -1 then return true end
	local link = GetInventoryItemLink("player", ContainerIDToInventoryID(bagId))
	if not link then return false end
	local linkId = select(3, link:find("item:(%d+)"))
	if not linkId then return false end
	local bagType = self:GetItemSubtype(linkId)
	if bagType and bagType == L["Bag"] then return true end
	return false
end

function ClosetGnome:LocateFreeSlot()
	for theBag = NUM_BAG_FRAMES, 0, -1 do
		if self:IsNormalBag(theBag) then
			if not slotLocks then slotLocks = {} end
			if not slotLocks[theBag] then slotLocks[theBag] = {} end
			local numSlot = GetContainerNumSlots(theBag)
			for theSlot = 1, numSlot, 1 do
				if not slotLocks[theBag][theSlot] then
					local texture = GetContainerItemInfo(theBag, theSlot)
					if not texture then
						return theBag, theSlot
					end
				end
			end
		end
	end
	return nil
end

function ClosetGnome:EquipItem(slot, item, secondTry)
	if item == false then
		-- We need to find a free slot in the inventory for this item.
		-- If there's anything in the slot.
		local hasItem = self:ItemNameFromSlot(slot)
		if hasItem ~= false then
			if not deequipQueue then deequipQueue = {} end
			table.insert(deequipQueue, slot)
		end
	else
		local bagNum, slotNum = self:FindItem(item)
		if bagNum > -1 then
			PickupContainerItem(bagNum, slotNum)
			if slot < 11 or slot == 15 or slot == 19 then
				AutoEquipCursorItem()
			else
				EquipCursorItem(slot)
			end
		else
			if not secondTry and item then
				self:ScheduleEvent("cg"..slot..item, self.EquipItem, 0.1, self, slot, item, true)
			else
				self:Print(L["Couldn't find %s in your inventory."]:format(tostring(item)))
			end
		end
	end
end

function ClosetGnome:HasSet(name)
	return self.db.char.set[name] ~= nil
end

function ClosetGnome:FindItem(item)
	for i = NUM_BAG_FRAMES, 0, -1 do
		for j = GetContainerNumSlots(i), 1, -1 do
			local link = GetContainerItemLink(i, j)
			if link and item == link then return i, j end
		end
	end
	return -1
end

-------------------------------------------------------------------------------
-- Hooks                                                                     --
-------------------------------------------------------------------------------

-- Only hooked if we are showing the paper doll frame with set adding
function ClosetGnome:CharacterFrame_OnHide()
	for id, item in pairs(slots) do
		if item.frame then
			item.frame:EnableMouse(false)
			item.frame:Hide()
			item.texture:Hide()
		end
	end

	_G.StaticPopup_Hide("ClosetGnomeAdd")

	if self:IsHooked("CharacterFrame_OnHide") then
		self:Unhook("CharacterFrame_OnHide")
	end
	if self:IsHooked("CharacterFrameTab_OnClick") then
		self:Unhook("CharacterFrameTab_OnClick")
	end
	if self:IsHooked("PaperDollItemSlotButton_OnModifiedClick") then
		self:Unhook("PaperDollItemSlotButton_OnModifiedClick")
	end
	if _G.Fizzle and self:IsHooked(_G.Fizzle, "UpdateItems") then
		self:Unhook(_G.Fizzle, "UpdateItems")
		if _G.Fizzle.db.profile.Border then
			_G.Fizzle:UpdateItems()
		end
	end

	_G.CloseAllBags()
	_G.HideUIPanel(_G.CharacterFrame)
	_G.CharacterFrame_OnHide()
end

function ClosetGnome:Noop() --[[ noop ]] end

-------------------------------------------------------------------------------
-- Events                                                                    --
-------------------------------------------------------------------------------

function ClosetGnome:AceEvent_FullyInitialized()
	-- Loop through the current sets and check if we have any of them equipped.
	for k,v in pairs(self.db.char.set) do
		local ret = self:IsSetFullyEquipped(k)
		if ret then
			lastEquippedSet = k
			self:TriggerEvent("ClosetGnome_WearSet", k)
			break
		end
	end
	self:UpdateDisplay()

	if not self.db.profile.hideBetaNotice then
		self:Print("|cff00ff00ClosetGnome is beta, but I consider it usable. Please direct comments to IRC or the forum. See the readme.txt for how to turn off this message.|r")
	end
end

function ClosetGnome:PLAYER_REGEN_ENABLED()
	if type(queuedSet) ~= "table" then
		self:UnregisterEvent("PLAYER_REGEN_ENABLED")
		queuedSet = {}
		return
	end
	self:WearSet(unpack(queuedSet))
end

-------------------------------------------------------------------------------
-- FuBarPlugin                                                               --
-------------------------------------------------------------------------------

function ClosetGnome:OnTextUpdate()
	if lastEquippedSet then
		self:SetText(self:GetSetDisplayName(lastEquippedSet))
	else
		self:SetText(L["ClosetGnome"])
	end
	if self.db.char.icons[lastEquippedSet] then
		self:SetIcon(self.db.char.icons[lastEquippedSet])
	else
		self:SetIcon(defaultIcon)
	end
end

function ClosetGnome:OnDoubleClick()
	if IsShiftKeyDown() then
		self:CreateOrUpdateSet(lastEquippedSet)
	else
		self:CreateOrUpdateSet()
	end
end

-- Hook this method and return true for any sets you do not want listed in the
-- tablet.
function ClosetGnome:HideSetFromUI(set)
	return false
end

-- Hook to provide for example coloring of different sets in the tablet.
function ClosetGnome:GetSetDisplayName(set)
	return set
end

local function clickfunc(name)
	if IsControlKeyDown() then
		local set = ClosetGnome.db.char.set[name]
		DressUpModel:Undress()
		for i = 1, 19 do
			local item = set[i]
			if i ~= 18 and item ~= false then
				DressUpItemLink(item)
			end
		end
	elseif IsShiftKeyDown() then
		ClosetGnome:CreateOrUpdateSet(name)
	else
		ClosetGnome:WearSet(name)
	end
end

function ClosetGnome:OnTooltipUpdate()
	local cat = nil

	local sets = {}
	for k in pairs(self.db.char.set) do table.insert(sets, k) end
	table.sort(sets)

	for i, set in ipairs(sets) do
		if not self:HideSetFromUI(set) then
			if not cat then
				cat = tablet:AddCategory(
					"columns", 2,
					"text", L["Available sets"],
					"showWithoutChildren", false,
					"child_hasCheck", true,
					"child_checked", true,
					"child_text2R", 0.5,
					"child_text2G", 0.5,
					"child_text2B", 0.5
				)
			end
			local displayName = self:GetSetDisplayName(set)
			if lastEquippedSet == set then
				displayName = L["%s |cffd9d919*|r"]:format(self:GetSetDisplayName(set))
			end
			cat:AddLine(
				"text", displayName,
				"func", clickfunc,
				"arg1", set,
				"checkIcon", self.db.char.icons[set] or "",
				"text2", self.db.char.keybindings[set] or ""
			)
		end
	end
	sets = reclaim(sets)

	if cat ~= nil then
		tablet:SetHint(L["|cffeda55fClick|r a set above to change your equipment. |cffeda55fCtrl-Click|r a set to try it on. |cffeda55fShift-Click|r a set to edit it. The last equipped set is indicated by a star."])
	else
		tablet:SetHint(L["You have not created any equipment sets yet. |cffeda55fDouble-Click|r to create a set."])
	end
end

